#include <string.h>
#include <stdio.h>
#include <math.h>
#include "defines.h"
#include "trd.h"

#include "exception_t.h"

void trd_read_sectors(BYTE *trd, unsigned start_track, unsigned start_sector, unsigned sector_count, BYTE *output)
{
	unsigned offset, to_read;
	struct exception e;
	
	if(start_track > 159 || start_sector > 15)
	{
		e.msg="Track/sector out of disc bounds";
		Throw e;		
	}
	
	offset = (start_track << 12) + (start_sector << 8);
	to_read = sector_count*0x100;
	if((offset+to_read) > TRD_SIZE)
	{
		e.msg="Track/sector out of disc bounds";
		Throw e;		
	}
		
	if(to_read) memcpy(output,trd+offset,to_read);
	return;
}

void trd_write_sectors(BYTE *trd, unsigned start_track, unsigned start_sector, unsigned sector_count, BYTE *data)
{
	unsigned offset, to_write;
	struct exception e;
		
	if(start_track > 159 || start_sector > 15)
	{
		e.msg="Track/sector out of disc bounds";
		Throw e;		
	}
	
	offset = (start_track << 12) + (start_sector << 8);
	to_write = sector_count*0x100;
	if((offset+to_write) > TRD_SIZE)
	{
		e.msg="Track/sector out of disc bounds";
		Throw e;		
	}		
		
	if(to_write) memcpy(trd+offset,data,to_write);
	return;
}

void trd_read(char *fname, BYTE *trd_buff)
{
	FILE *fp;
	struct exception e;
	
	memset(trd_buff,0,TRD_SIZE);
	
	if(NULL == (fp = fopen(fname,"rb")))
	{
		e.msg="Cannot open trd file for reading";
		Throw e;		
	}
	
	fread(trd_buff,1,TRD_SIZE,fp);
	fclose(fp);
	return;
}

void trd_write(char *fname, BYTE *trd_buff)
{
	FILE *fp;
	struct exception e;
	
	if(NULL == (fp = fopen(fname,"wb")))
	{
		e.msg="Cannot open trd file for writing";
		Throw e;		
	}
	
	if(TRD_SIZE != fwrite(trd_buff,1,TRD_SIZE,fp))
	{
		fclose(fp);
		e.msg="Cannot write data to trd file";
		Throw e;		
	}
	
	fclose(fp);
	return;	
}

int trd_fname_equal(struct trd_fname *fname1, struct trd_fname *fname2)
{
	if((!(fname1->name[0]) || fname1->name[0] == 0x01) && (!(fname2->name[0]) || fname2->name[0] == 0x01))
	{
		if(memcmp(&(fname1->name[1]),&(fname2->name[1]),7)) return(0);
		if(fname1->ext != fname2->ext) return(0);
		return(1);
	}
	
	return(!memcmp(fname1,fname2,sizeof(struct trd_fname)));
}

void trd_parse_fname(char *filename, char fname_ext_separator, char with_del_files, struct trd_fname *fname)
{
	char *sep;
	struct exception e;
	int len;
	
	if(!*filename || (len=strlen(filename)) > 10)
	{
		e.msg="Invalid tr-dos filename given";
		Throw e;	
	}
	
	memcpy(fname->name,"        ",8);
	fname->ext=' ';
	sep=strrchr(filename,fname_ext_separator);
	if(sep != NULL && *(sep+1))
	{
		fname->ext=*(sep+1);
		len=sep-filename;
	}
	
	memcpy(fname->name,filename,(len > 8)? 8 : len);
	
	if(with_del_files && *filename == with_del_files) fname->name[0]=0x01;
	
	return;
}

/*find catalogue entry for file. returns NULL if not found*/
struct trd_cat_entry *trd_find_cat_entry(BYTE *trd, struct trd_fname *fname, char with_del_files, int *is_last_file)
{
	struct trd_cat_entry *pcat;
	int i;
	
	for(i=0, pcat=(struct trd_cat_entry *)trd; i< 128; ++i, ++pcat)
	{
		if(((!(pcat->filename.name[0]) || pcat->filename.name[0] == 0x01) && !with_del_files)) continue;
			
		if(trd_fname_equal(fname, &(pcat->filename))) 
		{
			if(!(((pcat+1)->filename).name)[0] || i == 127) *is_last_file=1;
			else *is_last_file=0;
				
			return(pcat);
		}
	}	
	
	return(NULL);
}

void trd_writefile(BYTE *trd, BYTE *bfile_buff, int num_bytes, struct trd_fname *fname)
{
	int i, sector=0, track=0, num_sectors=0;
	struct trd_cat_entry *pcat = NULL;
	struct exception e;
	
	for(i=0, pcat=(struct trd_cat_entry *)trd; i< 128; ++i, ++pcat)
	{
		if(!(pcat->filename.name[0])) break;
			
		sector = ((pcat->start_sector + pcat->sectors) & 0x0F);
		track = pcat->start_track + ((pcat->start_sector + pcat->sectors) >> 4);
	}			
	
	if(i >= 128)
	{
		e.msg="File limit of 128 reached";
		Throw e;			
	}
	
	num_sectors = ceil((float)num_bytes / 0x100);
	
	if(num_sectors > 0x100)
	{
		e.msg="Too many sectors";
		Throw e;
	}		
	
	if(num_sectors > *(trd+(0x100*8)+0xE5))
	{
		e.msg="Not enough free sectors on disc";
		Throw e;
	}
	
	trd_write_sectors(trd, track, sector, num_sectors, bfile_buff);
	
	/*form trd_cat_entry*/
	
	memcpy(&(pcat->filename), fname, sizeof(struct trd_fname));
	
	pcat->sectors = num_sectors;
	pcat->start_sector = sector;
	pcat->start_track = track;
	/*TODO: support for BASIC program length*/
	pcat->length_l = (num_bytes & 0xFF);
	pcat->length_h = (num_bytes >> 8);
	
	/*update 8th sector of 0th track*/
	*(trd+(0x100*8)+0xE2) += ((*(trd+(0x100*8)+0xE1) + pcat->sectors) >> 4); /*first free track*/
	*(trd+(0x100*8)+0xE1) = (*(trd+(0x100*8)+0xE1) + pcat->sectors) & 0x0F; /*first free sector*/
	*(trd+(0x100*8)+0xE5) -= pcat->sectors; /*num. of free sectors*/
	++*(trd+(0x100*8)+0xE4); /*num. of files incl. deleted*/

	return;
}

void trd_deletefile(BYTE *trd, struct trd_cat_entry *file, int is_last_file)
{
	struct exception e;
	
	if(file->filename.name[0] && file->filename.name[0]!=0x01)
	{
		file->filename.name[0] = (is_last_file? 0x00: 0x01);
	
		/*update sys.sector(sec 8)*/
		if(is_last_file)
		{
			*(trd+(0x100*8)+0xE1) = file->start_sector; /*first free sector*/
			*(trd+(0x100*8)+0xE2) = file->start_track; /*first free track*/
			*(trd+(0x100*8)+0xE5) += file->sectors; /*num. of free sectors*/
			--*(trd+(0x100*8)+0xE4); /*num. of files incl. deleted*/
		}
		else ++*(trd+(0x100*8)+0xF4); /*num. of deleted files*/
		
		return;
	}
	
	e.msg="No such file to delete";
	Throw e;
}

